/*
  Copyright (C) TomasVlad 2012

  Permission to use, copy, modify, distribute, and sell this
  software and its documentation for any purpose is hereby granted
  without fee, provided that the above copyright notice appear in
  all copies and that both that the copyright notice and this
  permission notice and warranty disclaimer appear in supporting
  documentation, and that the name of the author not be used in
  advertising or publicity pertaining to distribution of the
  software without specific, written prior permission.

  The author disclaim all warranties with regard to this
  software, including all implied warranties of merchantability
  and fitness.  In no event shall the author be liable for any
  special, indirect or consequential damages or any damages
  whatsoever resulting from loss of use, data or profits, whether
  in an action of contract, negligence or other tortious action,
  arising out of or in connection with the use or performance of
  this software.
*/

/*****************************************************
Chip type               : ATmega8
Program type            : Application
AVR Core Clock frequency: 5,000000 MHz
Memory model            : Small
External RAM size       : 0
Data Stack size         : 256
*****************************************************/

#include <mega8.h>
#include <delay.h>

// #define LCD1602A    // for best PCB routing LCD 1602A

#define ulong unsigned long
#define uint unsigned int

#define LCD_CMD  0
#define LCD_DATA 1

#define LCD_E PORTB.0
#define LCD_RS PORTB.1

#ifdef LCD1602A
    #define LCD_D4 PORTB.5
    #define LCD_D5 PORTB.4
    #define LCD_D6 PORTB.3
    #define LCD_D7 PORTB.2
#else
    #define LCD_D4 PORTB.2
    #define LCD_D5 PORTB.3
    #define LCD_D6 PORTB.4
    #define LCD_D7 PORTB.5
#endif    

#define MODE_NULL PIND.6
#define MODE_CX	  PIND.7
#define MODE_SET  !PINC.1

#define CAL PORTD.5

#define RE  PORTD.0
#define DIR_RE DDRD.0 
#define DIR_IN  0
#define DIR_OUT 1

#define RX  PORTC.4
#define DIR_RX DDRC.4 

void msr_Uesr(void);
void msr_ElCm(void);

#pragma regalloc+
volatile uint Uesr_cnt;
volatile char cnt_TMR1_OVR;
volatile char cnt_TMR0_OVR;
volatile uint cnt_TMR1;
char Delay_3T @5;
uint Uesr @6;
#pragma regalloc-
// table
uint L0;
uint C0;
uint Kesr0;
uint Uesr0;
uint Re0;
ulong Cp0;
uint Ccal0;
uint Rcal;
uint KUesr;

char eeprom *EE_str_off;
char lcd_LineN;
char str_pos_now;
float F1pwr2;
float F0pwr2;
float FXpwr2;
ulong Cx;
uint Fx;
uint dFx;

uint ElCm_time;
uint Ucm0;
uint Ucm1;
uint UcmTop;
uint Utmp;
uint dUcm;

char dUc_200mS;
char dUc_100mS;
char lcd_value_point;
char tmp_idx = 0;
char str_pos_end;
char lcd_value_type;
uint Kesr;
char Uesr_OVR;
char graph;
ulong lcd_data;

// flags_mode
bit fmInitTmr;
bit fmRdyOk;
bit fmElCx;
bit fmLCx;
bit fmRe;
bit fmFreqStab;
bit fmIc;
bit fmRx;

// flags_stat
bit fsDelay_1T;
bit fsDelay_2T;
//bit fsPos7;
bit fsFreqChange;
bit fsMinus;
bit fsPos7;
bit fsCgAdr0;
bit fsCgAdr1;

#pragma keep+
eeprom char conf_lcd[] =    { 0x28,  // 4-bit interface, dual line, 5x7 dots
                            0x8,    // display off
                            0xC,    // display on, cursor off
                            0x1,    // clear display
                            0x6,    // display shift off, inc cursor move dir
                            0xFF };
eeprom char a_Cx[] =        {' ',0x5,' ','C','x',0xFF };
eeprom char a_Cp[] =        {' ',0x5,' ','C','p',0xFF };
eeprom char a_vC0[] =       {' ',0x5,' ','C','0',0xFF };
eeprom char a_Lx[] =        {' ',0x6,' ','L','x',0xFF };
eeprom char a_vL0[] =       {' ',0x6,' ','L','0',0xFF };
eeprom char a_ElCm[] =      {0x3,' ','C',' ','m',0xFF };
eeprom char a_rESR[] =      {0x4,' ','E','S','R',0xFF };
eeprom char a_Ccal[] =      {0x3,'C','c','a','l',0xFF };
eeprom char a_Rcal[] =      {0x4,'R','c','a','l',0xFF };
eeprom char a_KUesr[] =     {'K','U','e','s','r',0xFF };
eeprom char a_F0[] =        {' ','F','0',' ',' ',0xFF };
eeprom char a_Fcal[] =      {' ','F','C','a','l',0xFF };
eeprom char a_ElCsz[] =     {0x3,'C',' ','s','z',0xFF };
eeprom char a_rRx[] =       {' ',0x4,' ','R','x',0xFF };
eeprom char a_lcLx[] =      {0x5,0x6,' ','L','x',0xFF };
eeprom char a_Uesr0[] =     {'U','e','s','r','0',0xFF };
eeprom char a_Uesr[] =      {' ','U','e','s','r',0xFF };
eeprom char a_Fesr[] =      {' ','F','e','s','r',0xFF };
eeprom char a_icEsr[] =     {0x2,'-','E','S','R',0xFF };
eeprom char a_Re[] =        {' ',0x4,' ','R','e',0xFF };
eeprom char a_Ro[] =        {' ',0x4,' ','R','o'};
eeprom char a_0xFF[] =      {0xFF };
eeprom char a_C0[] =        {' ',' ',' ','C','-','-','-','>','0',' ',' ',' ',' ',' ',' ',' ',0xFF };
eeprom char a_L0[] =        {' ',' ',' ','L','-','-','-','>','0',' ',' ',' ',' ',' ',' ',' ',0xFF };
eeprom char a_Err[] =       {' ',' ',' ','>','>','!','E','r','r','o','r','!','<','<',' ',' ',0xFF };
eeprom char a_SwTo[] =      {' ',' ',' ',' ','S','w','i','t','c','h',' ','t','o',' ',' ',' ',0xFF };
eeprom char a_ToMsr[] =     {' ',' ','M','e','a','s','u','r','e',' ','M','o','d','e','!',' ',0xFF };
eeprom char a_ToCx[] =      {' ',' ',' ',' ',' ','C','x',' ','M','o','d','e','!',' ',' ',' ',0xFF };
eeprom char a_ban0[] =      {'L','C','M','3','-','M','E','G','A','8',' ','C','L','O','N','E',0xFF };
eeprom char a_ban1[] =      {'v','.','1','.','3','d',' ','(','c',')','T','o','m','a','s','V',0xFF };
eeprom char zgn_lcd[] =     { 0x0, 0x0, 0x3, 0x4, 0x2, 0x5, 0x9, 0x6,               // "delta"
                              0x18, 0x8, 0x18, 0xB, 0x1A, 0x3, 0x1, 0x3,            // "3/5"
                              0x0, 0x0, 0x13, 0x14, 0x14, 0x14, 0x13, 0x0,          // "ic"
                              0x4, 0x1F, 0x11, 0x1F, 0x0, 0x1F, 0x4, 0x0,           // "electrolit C"
                              0x4, 0xE, 0xA, 0xA, 0xA, 0xE, 0x4, 0x0,               // "Resistor"
                              0x4, 0x4, 0x1F, 0x0, 0x1F, 0x4, 0x4, 0x0,             // "Capicator"
                              0x8, 0x1D, 0x1D, 0x1D, 0x1D, 0x1D, 0x8, 0x0,          // "Inductor"
                              0x0, 0xE, 0x11, 0x11, 0x11, 0xA, 0x1B, 0x0, 0xFF };   // "Ohm"
eeprom char tbl_const[] =   { 0x00, 0x26,               // L0=9728 [pF*10]
                              0xCA, 0x17,               // C0=5322 [pF*10]
                              0xDD, 0x37,               // Kesr=
                              0x51, 0x0C,               // Uesr0=3153 [mV*10]
                              0xC2, 0x00,               // Re0=194 [Ohm]
                              0x2C, 0x06, 0x05, 0x00,   // Cp0=329260 [pF*10]
                              0x4D, 0x26,               // Ccal0=9805 [pF*10]
                              0xEA, 0x01,               // Rcal=490 [Ohm*10]
                              0x49, 0x01 };             // KUesr=329
                                                    
#pragma keep-

flash char range[] = { 'p','n','u','m',' ','k',0x0 };

interrupt [TIM0_OVF] void timer0_ovf_isr(void)
{
	if(fmInitTmr) {
		TCNT1 = 0;      // TMR1H=0;TMR1L=0;
		cnt_TMR1_OVR=0;
		TCNT0 = 0xC0;   // = -64
		fmInitTmr = 0;
	} else {
		cnt_TMR1 = TCNT1;
		cnt_TMR0_OVR++;
		
		if(cnt_TMR1_OVR > 7) {
			fmRdyOk=1;
			if((cnt_TMR1>>8) > 0) TIMSK &= ~(1<<TOIE1); // disable TMR1IE
		}
		
	}
}

interrupt [TIM1_OVF] void timer1_ovf_isr(void) { cnt_TMR1_OVR++; }

interrupt [EXT_INT1] void ext_int0_isr(void)
{
	char i;
    GICR &= ~(1<<INT1);     // disable INT1
    #asm("nop");
    #asm("nop");

    #asm 
    {
        MOV  R16,_Delay_3T
loop:
        SUBI  R16,1    // 1T
        BRNE loop   // 1/2T
        
        BST  R3,0   // 1T
        BRTC d2T    // 1/2T
        NOP
        NOP
d2T:
        BST  R3,1
        BRTC ddd
        NOP
        NOP
        NOP
ddd:
        NOP        
    }
    #endasm

    ADCSRA |= (1<<ADSC);    // (2T)ADC Start Conversion
	
    #asm("nop");
    #asm("nop");
    
    while((ADCSRA & (1<<ADIF))==0);     // wait ADSC==0
    ADCSRA |= (1<<ADIF);    // clear ADIF
    
	if(!(Uesr_cnt & 1)) {
		#asm
        {
            IN R30, 0x4
            IN R31, 0x4+1
            ADD R6, R30
            ADC R7, R31
        }
        #endasm

        MCUCR |= (1<<ISC11) | (1<<ISC10);   // The rising edge of INT1 generates an interrupt request
	} else {
//		Uesr -= ADCW;
		#asm
        {
            IN R30, 0x4
            IN R31, 0x4+1
            SUB R6, R30
            SBC R7, R31
        }
        #endasm

		MCUCR &= ~(1<<ISC10);   // The falling edge of INT1 generates an interrupt request.
	}
	
	Uesr_cnt++;
	
	if((Uesr & (1<<13)) && (Uesr & (1<<12))) {
	    Uesr_OVR++;
	    Uesr &= ~(1<<13);
    }

    GIFR |= (1<<INTF1);     // clear INTF1
    
    if(Uesr_cnt < 3200) GICR |= (1<<INT1);  // enable INT1 

    GIFR |= (1<<INTF1);     // clear INTF1

    i=0;
    while(i<64) {
        if(GIFR & (1<<INTF1)) break;
        i++;
    }

    GIFR |= (1<<INTF1);     // clear INTF1
}

void lcd_wrNybble(char n)
{
		LCD_D4 = (n & 1) ? 1 : 0;
		LCD_D5 = (n & 2) ? 1 : 0;
		LCD_D6 = (n & 4) ? 1 : 0;
		LCD_D7 = (n & 8) ? 1 : 0;
		delay_us(5);
		LCD_E=1;
		delay_us(5);
		LCD_E=0;
		delay_us(5);
}

void lcd_wrByte(char data, char flg)
{
	LCD_RS=0;
	delay_us(50);
	
	if(flg) LCD_RS=1;    // write data
	
	lcd_wrNybble((data >> 4) & 0xF);
	lcd_wrNybble(data & 0xF);
    delay_us(50);
	if(!str_pos_now) LCD_RS=0;
}

void lcd_EEdata_to_DDRAM(char eeprom *EE_adr)
{
	char tmp;
    char eeprom *adr;
	
	if(lcd_LineN != 0xFF) { 
		tmp = 0x80;
		if(lcd_LineN) tmp |= 0x40;	// lcd_line=1
		lcd_wrByte(tmp, LCD_CMD);   // set display line
	}

	adr=EE_adr;
	while(1) {
		tmp = *adr;
		if(tmp==0xFF) break;    // end data
		if(EE_adr==conf_lcd) lcd_wrByte(tmp, LCD_CMD);
		else lcd_wrByte(tmp, LCD_DATA);
		adr++;
	}	

	if(adr == EE_adr) {     // display clear line
		for(tmp=0; tmp<16; tmp++) lcd_wrByte(' ', LCD_DATA);	
	}
}

void lcd_show_EEstr(char eeprom *EE_Adr)
{
	lcd_EEdata_to_DDRAM(EE_Adr);
	delay_ms(100);
	if(str_pos_now) str_pos_now--;	
}

void msr_Freq(void)
{
	float tmp;
    uint adc_dly;

    
	delay_ms(100);
	FXpwr2 = 0;
    Fx = 0;
    fmInitTmr = 1;
	fmRdyOk = 0;
	cnt_TMR0_OVR = 0xFF;
	cnt_TMR1_OVR = 0;
	TCNT0 = 0xFE;
	TCNT1 = 0;
	TIFR |= (1<< TOV0); // T0IF = 0;
	// Enable interrupts
	TIMSK |= (1<<TOIE1);    // Timer1 Overflow Interrupt Enable
	TIMSK |= (1<<TOIE0);    // Timer0 Overflow Interrupt Enable bit
	#asm("sei")             // Global Interrupt Enable
	delay_ms(50);

	if(!fmInitTmr) {
		if(fmRe)
			while(!fmRdyOk && !(cnt_TMR1_OVR & 0x80));
		else
			while(!fmRdyOk && !(cnt_TMR1_OVR > 39));
	}

	// Disable interrupts
	#asm("cli")
	TIMSK &= ~(1<<TOIE0);    // disable Timer0 Overflow Interrupt
	TIMSK &= ~(1<<TOIE1);    // disable Timer1 Overflow Interrupt
	
	if(!fmRdyOk) return;

    tmp = 6400000.0*(float)(((ulong)cnt_TMR0_OVR<<2)+1);
    tmp /= (float)(((ulong)cnt_TMR1_OVR<<16) + cnt_TMR1);
	Fx = tmp;           // = F[kHz*10]
	FXpwr2 = tmp*tmp;  // = Fx^2[kHz^2*100]

	if(fmRe) FXpwr2 -= 0.6;
	
    Kesr = FXpwr2/55.0;    // = Fx^2/55.0
	
    tmp = (10000.0/(Fx*0.1))*1.75;
    adc_dly = (uint)(tmp - 69.0);

	Delay_3T = 0;
    while(adc_dly > 2) {
		adc_dly -= 3;
		Delay_3T++;
	}
	
	fsDelay_1T = adc_dly & 1;
    fsDelay_2T = (adc_dly >> 1) & 1;
}

void lcd_prep_str(void)
{
	char str[12];
	char pos;
	uint val, tmp_data;
	char idx, i;

	pos = 7;
    for(i=0;i<12;i++) str[i]=' ';
	str[7]='?';
    str[11]=0;
	
	if(fsPos7) str[2] = '=';
	else str[1] = '=';
	
	if((a_rESR == EE_str_off) && (fmRdyOk)) {
            str[0]=0;   // ==zgn_delta
            str[1]=1;   // ==zgn_graph
        }
	
	fsMinus = 0;
    if(lcd_data & (1<<31)) {
        fsMinus = 1;
        lcd_data = (uint)(0 - lcd_data);
    }
	
	if(lcd_LineN == 0) {
		tmp_data = lcd_data;
        if(fmElCx) idx = 0;
        else idx = 4;
start_shift:		
		if(tmp_data > 16) goto next_shift;
        if(idx > 6) goto end_shift;
next_shift:
			tmp_data >>= 1;
			idx++;
            goto start_shift;
end_shift:		
		idx -= 7;
		idx = (idx<<4)|(idx>>4);	// swap
		idx += (tmp_data & 0xFF);
		
        if(!fmRdyOk) idx=0xFF;
		tmp_idx += 2;
		tmp_idx -= idx;
		
		if(str_pos_now) {
			fsCgAdr0 ^= 1;
			if(!fsCgAdr0) fsCgAdr1 = 1;
			if(fsCgAdr0 && !fsCgAdr1) str_pos_now--;
		}
		
		if(tmp_idx >= 4) {
			tmp_idx = str_pos_end - str_pos_now;
			tmp_idx >>= 2;
			tmp_idx &= 0x3F;
			while(tmp_idx) {
				tmp_idx >>= 1;
				if(str_pos_end != 0xFF)	str_pos_end++;	
			}
			if(!str_pos_now) str_pos_now = str_pos_end;
		}
		
		tmp_idx = idx;
	}   // end lcd_LineN == 0		
		
	if(!fmRdyOk) goto copy_str;     // fmRdyOk==0

/* fmRdyOk == 1 */	
    while(lcd_data > 9743) {
        lcd_data /= 10;
        lcd_value_point++;
    }
		
    idx = 0;
	while(lcd_value_point > 2) {
        lcd_value_point -= 3;
        idx++;
    }
	
    // convert to dec	
	lcd_value_point += 4;
	if(fsPos7) lcd_value_point++;   // str[2] = '='
	if(lcd_value_point == 7) lcd_value_point = 0;
	
	// convert data to char's
	while(1) {
		if(pos == lcd_value_point) {    // init: pos==7
			str[pos] = '.';
            pos--;
			lcd_value_point = 0;
		}
		
		val = 0;
		while(lcd_data > 9) {
			lcd_data -= 10;
			val++;	
		}
		
		str[pos] = lcd_data + '0';
        pos--;
		
        if(fsPos7) {    // str[2] = '='
            pos = 7;
            fsPos7 = 0;                                                         
        }
		
        lcd_data = val;
		if(lcd_data) continue;
        if(!lcd_value_point) break;
	} 
	
	if(pos == 7) pos--;
	
	if(fsMinus) str[pos] = '-';
	
	str[9] = range[idx];     // idx='p','n','u','m',' ','k'
	
	if(dUc_100mS) { str[0] = dUc_200mS; str[1]= dUc_100mS; }
	
copy_str:    
    str[10] = lcd_value_type;
	dUc_100mS = 0;	

/* show LCD display data */
	lcd_EEdata_to_DDRAM(EE_str_off);
    for(i=0;i<11;i++) lcd_wrByte(str[i], LCD_DATA);
}

void cal_L0_C0(void)
{
    float sav_F1pwr2;
    float tmpC;
    
    sav_F1pwr2 = F1pwr2;
    F1pwr2 = FXpwr2;
    
    CAL = 1;
    delay_ms(500);
    msr_Freq();     // Fx=FreqNow; Kesr=(Fx^2)/55; FXpwr2=Fx^2 (fmRe=1 -> Fx^2-0.6)
    CAL = 0;

    F0pwr2 = FXpwr2;
    
    Cx = (ulong)((float)Ccal0/(F1pwr2/F0pwr2 - 1.0));    
    
    F0pwr2 = F1pwr2;
    
    if(!MODE_CX) F1pwr2 = sav_F1pwr2;
    
/* Lx=(1/(2*pi)^2)/(Cx*F0^2) */
    if(fmRe) tmpC = 2.533029591e+14F;
    else tmpC = 2.533029591e+15F;
    
    lcd_data = (ulong)(tmpC / ((float)Cx * F0pwr2));
    
    if(!MODE_CX) {      // mode==LX
        L0 = lcd_data;  // L0=Lx
        L0--;
        C0 = Cx;
        C0++;
    }
}

uint msr_DC_ADC(void)
{
    char i;
    uint res;
    
    delay_us(20);
    res = 3;

    for(i=0;i<49;i++) {
        ADCSRA |= (1<<ADSC);                // ADC Start Conversion
	    while((ADCSRA & (1<<ADIF))==0);     // wait ADIF==1
        ADCSRA |= (1<<ADIF);                // clear ADIF
        res += ADCW;
    }
    res &= 0xFFF8;
    return res;
}

void lcd_clear_line(char line)
{
    char tmp;
    
    tmp = lcd_LineN;
    lcd_LineN = line;
    lcd_EEdata_to_DDRAM(a_0xFF);
    lcd_LineN = tmp;
}

void init_lcm(void)
{
	char *ptr;
    char i;

	ADMUX &= 0xF0;  // ADC.0
    ADMUX |= 5;     // ADC.5(ElCx)
    CAL = 0;
    DIR_RE = DIR_OUT;
    RE = 0;
    DIR_RX = DIR_OUT;
    RX = 0;
    
    LCD_RS=0;
    LCD_E=0;
	
// ====== configure LCD ==========
    delay_ms(50);		// wait 16ms or more after power-on
    lcd_wrNybble(3);	// init
    delay_ms(5);
    lcd_wrNybble(3);	// init
    delay_ms(5);
	lcd_wrNybble(2);	// set 4-bit mode

/* Configure LCD interface */    
	lcd_EEdata_to_DDRAM(conf_lcd);
	delay_ms(10);		// 10mS

	lcd_wrByte(0x40,LCD_CMD);	    // Cmd=0x40(Sets CGRAM address 0x0)
	lcd_LineN=0xFF;     // no set LCD_LINE!

/* load LCD graphic's to CGRAM */
    lcd_EEdata_to_DDRAM(zgn_lcd); 
    
   	DIR_RE = DIR_IN;    // Re OFF
    
    lcd_LineN=0;
	lcd_EEdata_to_DDRAM(a_ban0);
    lcd_LineN=1;
	lcd_EEdata_to_DDRAM(a_ban1);

	delay_ms(2000);
    
    lcd_LineN = 0;
	str_pos_end = 64;
	str_pos_now = 0xFF;

	while(MODE_NULL) {
        lcd_show_EEstr(a_SwTo);
        lcd_LineN = 1;
        lcd_show_EEstr(a_ToMsr);
        lcd_LineN = 0;
    }	

	while(!MODE_CX) {
        lcd_show_EEstr(a_SwTo);
        lcd_LineN = 1;
        lcd_show_EEstr(a_ToCx);
        lcd_LineN = 0;
    }	

	str_pos_now = 64;

/* Load constants from EEPROM */
    ptr = (char*)&L0;
	for(i=0;i<20;i++) *(ptr+i) = tbl_const[i];

	delay_ms(2000);

    msr_Freq();     // Fx=FreqNow; Kesr=(Fx^2)/55; FXpwr2=Fx^2 (fmRe=1 -> Fx^2-0.6)
	EE_str_off = a_F0;  // ' ','F','0',' ',' '
	lcd_value_type = ' ';
	lcd_value_point = 17;
	fsPos7 = 0;         // str[1] = '='
	dUc_100mS = 0;
	lcd_data = Fx;
    lcd_prep_str();
    
    cal_L0_C0();
    lcd_data = Fx;
    lcd_LineN = 1;
    lcd_value_point = 17;
    EE_str_off = a_Fcal;
    lcd_prep_str();

    delay_ms(2000);
}

void msr_Uesr(void)
{
	Uesr_cnt = 0;
	Uesr = 15;
	dUc_100mS = 0;
	Uesr_OVR = 0;
	fmRdyOk=0;
	
	if(Fx < 650) return;   // < 65khZ
	if(Fx > 950) return;   // > 95khZ

	ADMUX &= 0xF0;  // ADC.0
    ADMUX |= 2;     //ADC.2(Uesr)
	delay_us(10);
    
	MCUCR &= ~(1<<ISC10);
    MCUCR |= (1<<ISC11);    // The falling edge of INT1 generates an interrupt request.
	GICR |= (1<<INT1);      // External Interrupt Request 1 Enable
 	
    #asm("sei");
    delay_ms(100);
	
	GICR &= ~(1<<INT1);     // External Interrupt Request 1 Disable
	#asm("cli");
	
	ADMUX &= 0xF0;  // ADC.0
    ADMUX |= 5;     // ADC.5(ElCx) 
	Uesr >>= 5;     // divide by 32
    Uesr &= 0x7FF;
    while(Uesr_OVR) { Uesr += 256; Uesr_OVR--; }
    
/*  ==Uesr[mV*10] */
	if(Uesr > 30) fmRdyOk=1;        // Uesr > 3.0[mV] 
}

void msr_Cp(void)
{
    char F_ReOn, F_ReOff;
    float sav_F1pwr2;
    
    sav_F1pwr2 = F1pwr2;  // save F1_pwr2
    fmLCx = 0;

//---- msr_Freq(ReOn[Low]) ---
    DIR_RE = DIR_OUT;
    RE = 0;
    fmRe = 1;
    msr_Freq();     // Fx=FreqNow; Kesr=(Fx^2)/55; FXpwr2=Fx^2 (fmRe=1 -> Fx^2-0.6)

//--- check Freq(ReOn[Low]) ---    
    F_ReOn = Fx & 0xFF;            // F_ReOn=Fx(ReOn);
    if(Fx >> 8) F_ReOn = 0xFF;     // Fx(ReOn) > 25.5[kHz]
    if(!F_ReOn) F_ReOn++;
    if(!fmRdyOk) F_ReOn = 0;

//--- msr_Freq(ReOff) ---
    DIR_RE = DIR_IN;
    fmRe = 0;
    msr_Freq();     // Fx=FreqNow; Kesr=(Fx^2)/55; FXpwr2=Fx^2 (fmRe=1 -> Fx^2-0.6)

//--- check Freq(ReOff) ---    
    F_ReOff = (Fx >> 3) & 0xFF;         // F_ReOff=Fx(ReOff)/8
    if((Fx >> 8) > 7) F_ReOff = 0xFF;   // Fx(ReOff) > 179.2[kHz]
    if(!F_ReOff) F_ReOff++;
    if(!fmRdyOk) F_ReOff = 0;
    
    if(!F_ReOn) goto cont;
    if(F_ReOn > 59) goto cont;       // F_ReOn > 5.9Khz
    if(F_ReOn > 23)                  // F_ReOn > 2.3kHz
        if(F_ReOn > F_ReOff) goto cont;  // F_ReOn > F_ReOff/8
    if(F_ReOn > 7) goto msr_ReOn;   // F_ReOn > 0.7kHz
    if(F_ReOff > 16) goto cont;     // F_ReOff > 12.8kHz

msr_ReOn:
//--- msr_Freq(ReOn[Low]) ---
    DIR_RE = DIR_OUT;
    RE = 0;
    fmRe = 1;
    msr_Freq();     // Fx=FreqNow; Kesr=(Fx^2)/55; FXpwr2=Fx^2 (fmRe=1 -> Fx^2-0.6)
    F_ReOff = 0;
    fmRdyOk = 1;

//-----------------------------------------    
cont:
    F0pwr2 = FXpwr2;  // F0pwr2=Fx^2(CcalOff && ReOn)
    
    if(F_ReOff) { fmRdyOk = 1; F_ReOn = 0; }    // F_ReOff > 0
    if(F_ReOn > 7) goto msr_CcalOn;             // F_ReOn > 0.7kHz 
    if(F_ReOff > 127) goto exit;                // F_ReOff > 101.6kHz
    if(!F_ReOff) goto exit;                     // F_ReOff ==0

/* !!! measure parasitic Cx !!! */    
msr_CcalOn:
    F1pwr2 = F0pwr2;  // F1_pwr2=F^2(CcalOff && ReOn)

//--- msr_Freq(CcalOn && ReOff) ---
    CAL = 1;        // Ccal=On
    delay_ms(500);
    msr_Freq();     // Fx=FreqNow; Kesr=(Fx^2)/55; FXpwr2=Fx^2 (fmRe=1 -> Fx^2-0.6)
    CAL = 0;        // Ccal=Off
    F0pwr2 = FXpwr2;  // F0pwr2=Fx^2(CcalOn && ReOff)
//--- calc Cx ---    
    Cx = (float)Ccal0 / (F1pwr2 / F0pwr2 - 1.0);
    fmLCx = 1;
    
exit:
    F1pwr2 = sav_F1pwr2;  // restore F1_pwr2
}

void set_Null(void)
{
    uint Ucal;
    char i;
    char *ptr;
    
    lcd_clear_line(1);
    lcd_LineN = 0;

    delay_ms(100);
    
    if(!fmRdyOk || MODE_NULL==0)
    {
        lcd_EEdata_to_DDRAM(a_Err);
        goto exit;
    }
    
    if(MODE_SET)
    {
set_ccal:
        if(!MODE_NULL)
            if((Ccal0 >> 8) < 46) Ccal0++;  // < 1177.6[pF]
        if(!MODE_CX)
            if((Ccal0 >> 8) > 34) Ccal0--;  // > 870.4[pf]
        
        lcd_value_type = 'F';
        lcd_data = Ccal0;
        EE_str_off = a_Ccal;    // ' ', 'C', 'c', 'a', 'l'
        lcd_value_point = 2;    // [pF]
        str_pos_now = str_pos_end;
        lcd_prep_str();
        delay_ms(100);
        if(!MODE_SET) goto set_rcal;
        goto set_ccal;

set_rcal:
        delay_ms(500);
        while(!MODE_SET);
set_rcal_:        
        if(!MODE_NULL)
            if(Rcal < 500) Rcal++;
        if(!MODE_CX)
            if(Rcal > 400) Rcal--;
        
        lcd_value_type = 7;     // "Ohm"
        lcd_data = Rcal;
        EE_str_off = a_Rcal;    // ' ', 'R', 'c', 'a', 'l'
        lcd_value_point = 14;
        str_pos_now = str_pos_end;
        lcd_prep_str();
        delay_ms(500);
        if(!MODE_SET) goto set_kuesr;
        goto set_rcal_;
        
set_kuesr:
        delay_ms(500);
        while(!MODE_SET);
set_kuesr_:        
        if(!MODE_NULL)
            if(KUesr < 700) KUesr++;
        if(!MODE_CX)
            if(KUesr > 100) KUesr--;
        
        lcd_value_type = ' ';
        lcd_data = KUesr;
        EE_str_off = a_KUesr;   // 'K', 'U', 'e', 's', 'r'
        lcd_value_point = 14;
        str_pos_now = str_pos_end;
        lcd_prep_str();
        delay_ms(200);
        if(!MODE_SET) goto end_set;
        goto set_kuesr_;
    }

end_set:    
    DIR_RE = DIR_OUT;
    RE = 1;
    delay_us(500);
    
    Ucal = msr_DC_ADC();  // ADC.2(AN6)

    DIR_RE = DIR_IN;

#define URE  47120.0F      // [mV*10]   
    if((Ucal >> 8) > 27)   // > 691.2 mV
    {
//      Re0 = (URE/Ucal- 1.0)*Rcal/10.0;     // Re=(Rcal*U0)/Ux - Rcal
        Re0 = (uint)((URE/Ucal- 1.0)*Rcal/10.0);
    }
    
    lcd_clear_line(0);
    lcd_value_type = 7;   // "Ohm"
    lcd_value_point = 15;
    lcd_data = Re0;
    EE_str_off = a_Re;  //' ',' ',' ','R','e'
    fsPos7 = 0;         // str[1] = '='
    lcd_prep_str();

    delay_ms(2000);

    lcd_EEdata_to_DDRAM(a_SwTo);
    lcd_LineN = 1;
    lcd_EEdata_to_DDRAM(a_ToMsr);
    str_pos_now = 0xFF;
    
    while(MODE_NULL) {
        lcd_LineN = 0;
        lcd_show_EEstr(a_SwTo);
        lcd_LineN = 1;
        lcd_EEdata_to_DDRAM(a_ToMsr);
    }
    
    lcd_LineN = 0;
    str_pos_now = str_pos_end;

    delay_ms(2000);
    
    msr_Freq();     // Fx=FreqNow; Kesr=(Fx^2)/55; FXpwr2=Fx^2 (fmRe=1 -> Fx^2-0.6)
    
    if(Fx > 1050) {      // >105kHZ
        cal_L0_C0();
        fmElCx = 0;
        lcd_clear_line(0);
        if(MODE_CX) {
            lcd_value_type = 'F';
            lcd_value_point = 2;
            EE_str_off = a_vC0;
            lcd_data = Cx;
            lcd_prep_str();
            lcd_LineN = 1;
            lcd_EEdata_to_DDRAM(a_C0);  // 'C', '-', '>', '0', 0FF
        } else {
            lcd_value_type = 'H';
            lcd_value_point = 7;
            EE_str_off = a_vL0;
            lcd_prep_str();
            lcd_LineN = 1;
            lcd_EEdata_to_DDRAM(a_L0);  // 'L', '-', '>', '0', 0FF
        }
    } else {
        Kesr0 = Kesr;
        msr_Uesr();
        Uesr0 = Uesr;
        msr_Cp();       // res=Cx; set fmLCx!!!
        Cp0 = Cx;
        
        //show Ro0
        lcd_data = L0;
        lcd_data *= L0;
        lcd_data >>= 16;
        lcd_data *= Kesr;
        lcd_data /= Uesr0;
//        Ro0 = lcd_data;
        
        lcd_LineN = 0;
        lcd_value_type = 7;       // "Ohm"
        lcd_value_point = 11;
        EE_str_off = a_Ro;
        lcd_prep_str();

       // show Cp
//        lcd_clear_line(1);
        lcd_LineN = 1;
        lcd_value_type = 'F';
        lcd_value_point = 2;
        EE_str_off = a_Cp;  // ' ',' ',' ','C','p'
        lcd_data = Cx;
        lcd_prep_str();
        delay_ms(2000);
        
        // show Uesr    
        lcd_LineN = 0;
        lcd_value_type = 'V';
        lcd_value_point = 11;
        EE_str_off = a_Uesr0;
        fsPos7 = 0;     // str[1] = '='
        lcd_data = Uesr0;
        lcd_prep_str();
//        delay_ms(4000);
        
        lcd_LineN = 1;
        lcd_value_type = ' ';
        lcd_value_point = 17;
        EE_str_off = a_Fesr;
        lcd_data = Fx;
        lcd_prep_str();    
    }
    // write to EEPROM
    ptr = (char*)&L0;
	for(i=0;i<20;i++) tbl_const[i] = *(ptr+i);
    
exit:
    delay_ms(2000);
}

void msr_ElCm(void)
{
    uint Ux;
    
    ElCm_time = 0;
    lcd_data = 0;
    
    if(fmRx || fmIc) {
        delay_ms(500);
        return;
    }
    
    EE_str_off = a_ElCm;       // 3,' ','C',' ','m'
    fmElCx = 0;
    
    DIR_RE = DIR_OUT;
    RE = 0;
    delay_ms(500);          // discharge elCm 500mS    
    DIR_RE = DIR_IN;

    Ucm0 = msr_DC_ADC();    // ADC.2
        
    if(Ucm0 & (1<<15)) goto discharge_Cm_5mS;   // Ucm0 > 3276.8[mV]
        
    Ucm1 = (Ucm0 + 10000) / 49;     // Ucm1 = Ucm0 + 1000.0[mV]
        
    // charge Cm to Ucm0 + 1000.0[mV]
    DIR_RE = DIR_OUT;
    RE = 1;
    
    while(1) {
        ElCm_time++;
        ADCSRA |= (1<<ADSC);                // ADC Start Conversion
	    while((ADCSRA & (1<<ADIF))==0);     // wait ADSC==0
        ADCSRA |= (1<<ADIF);
        if((ElCm_time & (1<<14)) || (ADCW >= Ucm1)) break;
    }
        
    DIR_RE = DIR_IN;
    
    UcmTop = msr_DC_ADC();
    if((ElCm_time & (1<<14)) && ((Ucm0+256) > UcmTop)) fmRx = 1;
    if(ElCm_time < 2) goto discharge_Cm_5mS; 
    
    // ElCm connected    
    fmElCx = 1;
    fsPos7 = 1;     // str[2] = '='
    lcd_value_point = 6;
    lcd_value_type = 'F';
    Utmp = UcmTop;
        
    while(1) {
        delay_ms(100);
        Ux = msr_DC_ADC();    //ADC.2
        Utmp -= Ux;
        Utmp >>= 4;
        Utmp &= 0xFFF;
        dUc_200mS = 0;
        if((Utmp & (1<<11))) goto next;
        
        if(Utmp > 99) Utmp = 99;    // 158.4[mV]
            
        if(fmElCx) dUc_100mS = (char)Utmp;  // dT==100mS
        else {                              // dT==200mS
            dUc_200mS = (char)Utmp;
            if(dUc_200mS <= 1) dUc_200mS = 0;
            dUc_200mS <<= 1;
        }
next:
        fmElCx ^= 1;
        Utmp = Ux;
        if(fmElCx) break;            
    }
        
    // --- show el_C_sz ---
    if(dUc_100mS < dUc_200mS) EE_str_off = a_ElCsz;    // 3,'C',' ','s','z'
        
    dUc_200mS = '0';
    while(dUc_100mS > 9) { dUc_100mS -= 10; dUc_200mS++; }
    dUc_100mS += '0';
        
    if(dUc_200mS == '0') { dUc_200mS = dUc_100mS; dUc_100mS = ' '; }
        
    dUcm = UcmTop - Ucm0;
    
    // elC = [(50127-Ucm0)*23400/dUcm-12500]/Re*cnt-49
    // elC = (Ue/Re)/dUc*Tc

#define DCADC_TIME 24872UL  // ADC confersion time    
    lcd_data = (((((50127UL - Ucm0) * DCADC_TIME)
                / dUcm) - 12500UL) / Re0)
                * ElCm_time - 49UL;

    
    if(lcd_data < (3<<8)) {   // elC < 768[nF]
        lcd_data *= 10;
        lcd_value_point--;
    }
        
discharge_Cm_5mS:        
    DIR_RE = DIR_OUT;
    RE = 0;
    delay_ms(5);
    Utmp = msr_DC_ADC();
    DIR_RE = DIR_IN;
        
    if(!(ElCm_time & (1<<14))) return;
    if(Ucm0 >> 8) return;       // Ucm0 > 256
    if(!(dUcm >> 8)) return;    // dUcm <= 256
    if(Utmp >> 8) return;       // disUcm > 256
        
    if(fmFreqStab) { fmFreqStab = 0; fmRdyOk = 0; }
    else { fmRx = 0; fmIc = 1; }    // mode=Ic
}

void calc_Cx(void)
{
    lcd_data = (float)Cx * (F1pwr2/F0pwr2 - 1.0);
    EE_str_off = a_Cx;      // ' ',5,' ','C','x'
    lcd_value_point = 2;
    lcd_value_type = 'F';
}

void msrCx_calcLx(void)
{
    float k;
    
    lcd_value_type = 'H';
    lcd_value_point = 7;
    
    if(fmFreqStab) {
        if(fmRdyOk) fmFreqStab = 0;
        msr_Cp();   // res=Cx; set fmLCx!!!
    }
    
    if(!fmLCx) {            // fmLCx==0 
        if(fmRe) Cx = Cp0;  // Re==On
        else Cx = C0;       // Re==Off
    }
/* calculate Lx=(1/(2*pi)^2)/(Cx*F0^2) */

    k = 2.533029591e+14F;
    if(!fmRe) k *= 10.0;
    
    lcd_data = (ulong)(k / ((float)Cx * F0pwr2));
    
    if(fmLCx) EE_str_off = a_lcLx;
    else EE_str_off = a_Lx;

    if(fmRe) {              // fmRe==1
        lcd_value_point++;
        fsPos7 = 1;         // str[2] = '='
    } else {                // fmRe==0
        delay_ms(500);      // ???????????
        lcd_data -= L0;     // = Lx-L0
    }
}

void msr_Rx(void)
{
    float Ro, dUesr, tUesr0;
    ulong tmp;
    char mask;
    
    fsMinus = 0;
    fsPos7 = 1;     // str[2] = '='
    
    tUesr0 = Uesr0;
    
    if(Kesr0 > Kesr) tUesr0 *= (float)Kesr0;
    else {
        tUesr0 *= (float)Kesr;
        Kesr = Kesr0;
    }
    
    tUesr0 /= (float)Kesr;
    
    msr_Uesr();         // res=Uesr
    
    if(!fmRdyOk) return;

    Ro = (float)L0;
    Ro *= Ro;
    Ro *= (float)Kesr;
    Ro /= (float)KUesr;
    Ro /= 100.0;
    Ro /= tUesr0;

    dUesr = tUesr0 - (float)Uesr;
    if(Uesr > Uesr0) {
        fsMinus = 1;
        dUesr = (float)Uesr - tUesr0;
    }

    Ro *= dUesr;

    if(fmIc) {
        tmp = ((ulong)Ro >> 16) / 3;
        if(Uesr < tmp) fmRdyOk = 0;
        Uesr -= tmp;
        if(Uesr < 29) fmRdyOk = 0;   
    }
    
    lcd_data = (ulong)(Ro/(float)Uesr);

    if(fsMinus) { lcd_data = 0 - lcd_data; return; }
    if(fmRx || fmIc) return;

/* ================ graph ElCx ESR ============================= */

    ElCm_time |= 1;
    tmp = lcd_data * ElCm_time;  // =Resr*ElCm
    tmp >>= 16;
    
    lcd_wrByte(0x48,LCD_CMD);     // Sets_CGRAM_address 0x8
    
    graph = 0;
    if(tmp & 0xFF00) graph = 0x1F;
        
    tmp &= 0xFF;
    mask = 0xFF;
    while(mask)
    {
        mask >>= 1;
        
        if(mask < (char)tmp) {
            graph <<= 1;
            graph |= 1;
        }
 
       graph &= 0x1F;
        lcd_wrByte(graph,LCD_DATA);
    }
}

void main(void)
{
// Port B initialization
PORTB=0x00;
DDRB=0b00111111;
// Port C initialization
// PC.4==RX
PORTC=0x00;
DDRC=0b00010000;
// Port D initialization
// PD.0=RE; PD.5=CAL
PORTD=0x00;
DDRD=0b00100001;
// Timer/Counter 0 initialization
// Clock source: T0 pin Falling Edge
TCCR0=0x06;
TCNT0=0x00;
// Timer/Counter 1 initialization
TCCR1A=0x00;
TCCR1B=0x01;
TCNT1H=0x00;
TCNT1L=0x00;
ICR1H=0x00;
ICR1L=0x00;
OCR1AH=0x00;
OCR1AL=0x00;
OCR1BH=0x00;
OCR1BL=0x00;
// Timer/Counter 2 initialization
ASSR=0x00;
TCCR2=0x00;
TCNT2=0x00;
OCR2=0x00;
// External Interrupt(s) initialization
MCUCR=0x00;
// Timer(s)/Counter(s) Interrupt(s) initialization
TIMSK=0x05;
// Analog Comparator initialization
ACSR=0x80;
SFIOR=0x00;
SFIOR |= (1<<PUD);  // Pull-up Disable
// ADC initialization
ADMUX = (1<<REFS0);     // Vref = AVCC with external capacitor at AREF pin
ADCSRA = (1<<ADPS2);    // Division Factor=16
ADCSRA |= (1<<ADEN);    // ADC Enable 

    init_lcm();

    while(1)
    {
        msr_Freq();     // Fx=FreqNow; Kesr=(Fx^2)/55; FXpwr2=Fx^2 (fmRe=1 -> Fx^2-0.6)
        F0pwr2 = FXpwr2;

//        dFx += 1.0;
        if(dFx > Fx) dFx -= Fx;
        else dFx = Fx - dFx;
        
        if(dFx > 152) fsFreqChange = 1;
        else {
            if(fsFreqChange) fmFreqStab = 1;
            fsFreqChange = 0;
        }
        
        dFx = Fx;

        if(Fx > 1050) {  // Fx > 105kHz
            fmRx = 0;
            fmIc = 0;
            fmElCx = 0;
        }
        
        if(MODE_NULL) set_Null();
        
        if(MODE_CX)
        {
            if(fmRe) {
                DIR_RE = DIR_IN;
                fmRe = 0;
            }
            
            DIR_RX = DIR_IN;
            fmLCx = 0;
            msr_ElCm();
            DIR_RX = DIR_OUT;
            RX = 0;
            
            if(!fmElCx) calc_Cx();
        } else msrCx_calcLx();
        
        lcd_LineN = 0;
        if(!fmRx && !fmIc) lcd_prep_str();  // lcd_LineN==0
        
        if(MODE_NULL) set_Null();

        lcd_LineN = 1;

        if(fmElCx) {
            msr_Rx();
            lcd_value_type = 7;       // "Ohm"
            lcd_value_point = 11;
            EE_str_off = a_rESR;    // 4,' ','E','S','R'
            
            if(fmRx) {      // show "Rx"
                EE_str_off = a_rRx;  // ' ',4,' ','R','x'
                lcd_LineN = 0;
            }
            
            if(fmIc) {      // show "r-ESR"
                EE_str_off = a_icEsr;  // 2,'-','E','S','R'
                lcd_LineN = 0;
            }
            
            lcd_prep_str();
            
            if(fmRx) {      // show "Uesr"
                lcd_LineN = 1;
                lcd_value_type = 'V';
                lcd_value_point = 11;
                EE_str_off = a_Uesr;
                lcd_data = Uesr;
                lcd_prep_str();
            }
            goto is_lcd_LineN;
        }
        
        if(fmLCx) {     // show "Cp"
            lcd_value_type = 'F';
            lcd_value_point = 2;
            EE_str_off = a_Cp;  // ' ',' ',' ','C','p'
                 
            if(fmRe) {
                fsPos7 = 1;     //str[2] = '='
                lcd_data = Cx - Cp0;
                if(lcd_data < 63) fmLCx = 0;
            } else lcd_data = Cx - C0;
            
            if(lcd_data & (1<<31)) fmLCx = 0;
            if(fmLCx) lcd_prep_str();
            goto is_lcd_LineN;
        }

        lcd_LineN = 0;

is_lcd_LineN:
        if(!lcd_LineN) lcd_clear_line(1);   // lcd_LineN==0        
    };
}
